/* schedprt7 */
#define VERSION "7.0"

/*  V1.0 Initial release at the end of 2006.
    V1.1 - Jul 2007 Fixed bug where songs that exactly fitted one page of the Generic/Text
	Only Printer caused the author and copyright information to be appended to the
	song. A chunk of subroutine get_song was re-written.
    V1.2 - 7 Jan 2012 added a running tally of song to help with diagnostics.
    V7.0 - a major rework to cater for EasyWorship v7
*/

/* #define UNIX	/* Uncomment this line to compile on a unix/linux system */
				/* comment it out to compile on a Windows system */

/* Compile with
	gcc -o schedprt7 twrapper.c schedprt7.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define NULLCH (char)0

void usage(void)
{
	fprintf(stderr,"Usage (v%s): schedprt [-c config.dat] [-o output.rtf] <filename.txt> -[h|?]\n",
		VERSION);
}

/* Copyright (c) 2006-25 Rick Parsons.

	This software is published Open Source Freeware. You are free to use the
	programs either in full or part without charge. I would ask however that
	this notice is included with any extract or onward distribution.

	No charge may be made for the use of this code, however there is no
	objection to including modest extracts (with acknowledgement) in otherwise
	commercial software.

	No warranty is given or implied by the use of the software which you do at
	your own risk. No compensation can be considered regarding damage to data,
	computers or any thing else arising as a result of using these programs.
*/

#define TRUE 1
#define FALSE 0

int options(int, char **);
void init(void);
void get_template(FILE *);
void get_tname(FILE *, char *);
void get_line(FILE *, char *);
void get_escape(FILE *, char *);
char * get_song(FILE *);
char * get_stanza(void);
void put_schedule(FILE *, FILE *);
void put_text(FILE *, char *, FILE *);
void store_escape(int, char *);
int is_tag(char *);
int tgetc(FILE *);
int ugetc(FILE *);

void splitpath(const char *, char *, char *, char *);
char * makepath(const char *, const char *, const char *);

/* escape = an array of char *, one for each possible escape a-z, A-Z, 0-9
		(assumes numbers before upper case before lower case)
   output = output file name
   config = config file name
   songbuf = buffer for one song
   temp = a buffer for various purposes
   inschd = a flag which is set while processing schedule content
   insong = a flag which is set while processing song content
	These are to guard against some indefinate loops
   sbp = a pointer into songbuf to mark the next stanza
   tags = a table of valid stanza tags
*/

#define ESCAPECHARS 'z'-'0'+1
static char * escape[ESCAPECHARS];
static char output[128];
static char config[128];
#define SONGBUFSZ 2048
static char songbuf[SONGBUFSZ];
static char temp[SONGBUFSZ];
static int inschd, insong;
static char * sbp;
static char * tags[] = {"Verse", "Chorus", "Bridge", "Intro", "End", "Slide", "Tag", ""};

/* fixed meaning escape codes
	: = Schedule template
	; = song template
	< = end of line template
	0 = stanza template
	1-7 = tag templates as above
	A = Author
	C = schedule Content (special coding)
	L = Licence
	N = song Number
	S = Schedule name
	T = Title
	a = Reserved (administrator data)
	c = copyright
	l = user assignable licence
	n = tag number/text
	s = song content (special coding)
	t = tag
	z = stanza content
	the rest in the gaps are not used directly but can be defined in the config file
*/
#define ESCAPE_schedule ':'-'0'
#define ESCAPE_song ';'-'0'
#define ESCAPE_line '<'-'0'
#define ESCAPE_stanza '0'-'0'
#define ESCAPEA 'A'-'0'
#define ESCAPEC 'C'-'0'
#define ESCAPEL 'L'-'0'
#define ESCAPEN 'N'-'0'
#define ESCAPES 'S'-'0'
#define ESCAPET 'T'-'0'
#define ESCAPEa 'a'-'0'
#define ESCAPEc 'c'-'0'
#define ESCAPEl 'l'-'0'
#define ESCAPEn 'n'-'0'
#define ESCAPEs 's'-'0'
#define ESCAPEt 't'-'0'
#define ESCAPEz 'z'-'0'

#define BUFSZ 1024

int _main(int argc, char *argv[])
{
	extern char output[];
	extern char config[];
	int i;
	char input[128]; /* input file name */
	char path[128];
	char fname[128];
	char ext[128];
	int c;
	char * cp;
	FILE * in;
	FILE * out;
	FILE * cfg;

	/* preset some variables */
	*output = NULLCH;
	*config = NULLCH;

	/******************************/
	/* Decode the call parameters */
	/******************************/

	if (argc == 1)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	i = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (i == 0)
	{
		usage();
		return 2;
	}

	/* check bounds of input path before using */
	if (strlen(argv[i]) > 127)
	{
		fprintf(stderr, "Fatal: File path too long, move it closer to the root of the disk\n");
		return 12;
	}

	strcpy(input, argv[i]);

	if (*output == NULLCH)
	{
		/* if we have no output file name then create it from the input */
		splitpath(input, path, fname, ext);
		strcpy(output, makepath(path, fname, "rtf"));
	}

	if (*config == NULLCH)
	{
		/* if we have no config file name then use the default */
		/* look for it in the program path, not cwd */
		/* check size of path before using */
		if (strlen(argv[0]) > 126)
		{
			fprintf(stderr, "Fatal: Program path too long, move it closer to the root of the disk\n");
			return 12;
		}
		splitpath(argv[0], path, fname, ext);
		strcpy(config, makepath(path, "schedprt7", "dat"));
	}

	/* initialise some stuff, mostly the escape array */
	init();

	/* open the configuration file */
	if ((cfg = fopen(config, "r")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Configuration file %s failed to open\n", config);
		perror("Fatal");
		return 12;
	}

	/* read the configuration file */
	get_template(cfg);
	fclose(cfg);

	/* open the input file */
	if ((in = fopen(input, "r")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Input file %s failed to open\n", input);
		perror("Fatal");
		return 12;
	}

	/* Initialise the schedule name in advance of needing it */
	cp =  input;
	while (*(cp++) != '.')
		;
	*(--cp) = NULLCH;
	store_escape(ESCAPES, input);

	/* open the output file */
	if ((out = fopen(output, "w")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Output file %s failed to open\n", output);
		perror("Fatal");
		fclose(in);
		return 12;
	}

	/* skip over the leading code */
	c = ugetc(in);
	if (c != 0xFE)
	{
		fprintf(stderr, "Fatal: Unicode identifier missing\n");
		perror("Fatal");
		fclose(in);
		return 12;
	}
	
	/* This is where the bulk of the work is done */
	put_schedule(out, in);

	fclose(out);
	fclose(in);

	return 0;
}

int ugetc(FILE *stream)
{	/* read a potentially UTF-8 character - also ignore CR */
	/* Not very strict as it allows through single bytes up to 0xBF and doesn't
		check subsequent byte headers */
	int c, d, e, f, x;
	c = getc(stream);
	if (c == '\r')
		c = getc(stream);
	if (c > 0xC0)
	{
		d = getc (stream);
		if (c > 0xE0)
		{
			e = getc(stream);
			if (c > 0xF0)
			{
				f = getc(stream);
				/* 4 byte code */
				x = (c & 0x07) << 18 | (d & 0x3f) << 12 | (e & 0x3F) << 6 | (f & 0x3F);
			}
			else
				/* 3 byte code */
				/* fprintf(stderr, "char read %d %d %d\n", c, d, e); debug */
				x = (c & 0x0F) << 12 | (d & 0x3f) << 6 | (e & 0x3F);
		}
		else
			/* 2 byte code */
			/* fprintf(stderr, "char read %d %d\n", c, d); debug */
			x = (c & 0x1F) << 6 | (d & 0x3F);
	}
	else
		/* 1 byte code */
		x = c;

	/* now fake the unicode character into one byte */
	if (x < 0xFF)
		return x;

	c = x;
	/* fprintf(stderr, "char %d\n", c); debug */
	switch (c)
	{
		case 0x2019: x = 0x92; /* apostrophe */
			break;
		case 0x2028: x = '\n'; /* soft new line */
			break;
		case 0x2029: x = 0; /* P SEP */
			break;
		case 0xFEFF: x = 0xFE;
			break;
		default: x = '#';
	}
	return x;
}

int tgetc(FILE *stream)
{	/* read a character - also ignore CR */
	int c;
	c = getc(stream);
	if (c == '\r')
		c = getc(stream);
	return c;
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	extern char output[];
	extern char config[];
	int i,firstnonoption=0;

	for (i=1; i < argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'c':
				case 'C': /* Configuration file name */
					strcpy(config, argv[++i]);
					break;
				case 'o':
				case 'O': /* Output file name */
					strcpy(output, argv[++i]);
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				default:
					fprintf(stderr,"Unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = NULLCH;
	*ext = NULLCH;

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = NULLCH;
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = NULLCH;
	}

	if ((cp = strchr(file, '.')) != (char *)NULL)
	{
		*cp = NULLCH;
		strcpy(ext, cp + 1);
	}

	return;
}

char * makepath(const char * path, const char * file, const char * ext)
{
	extern char temp[];

	*temp = NULLCH;
	if (*path != NULLCH)
	{
		strcpy(temp, path);
#ifdef UNIX
		strcat(temp, "/");
#else
		strcat(temp, "\\");
#endif
	}
	strcat(temp, file);
	strcat(temp, ".");
	strcat(temp, ext);

	return temp;
}

void init(void)
{
	extern char * escape[];
	int i;

	/* initialise all escapes to point to a null string */
	for (i = 0; i < ESCAPECHARS; i++)
	{
		escape[i] = (char *)NULL;
		store_escape(i, "");
	}

	/* One has a pre set default value */
	store_escape(ESCAPE_line, "\\line\n");

	return;
}

void get_template(FILE * cfg)
{
	extern char * escape[];
	int i, esc;
	char c;

	/* fprintf(stderr, "get_template\n"); debug */

	while ((c = tgetc(cfg)) != EOF)
	{
		if (c != '%')
		{	/* Comment or other unknown text */
			if (c == '\n')
				continue;

			while ((c = tgetc(cfg)) != EOF)
				if (c == '\n')
					break;
			continue;
		}
		get_tname(cfg, temp);
		if (strcasecmp(temp, "Schedule") == 0)
			esc = ESCAPE_schedule;
		else if (strcasecmp(temp, "Song") == 0)
			esc = ESCAPE_song;
		else if (strcasecmp(temp, "Line") == 0)
			esc = ESCAPE_line;
		else if (strcasecmp(temp, "Stanza") == 0)
			esc = ESCAPE_stanza;
		else
		{
			/* see if this is a special tag template */
			for (i = 0; *tags[i] != NULLCH; i++)
			{
				if (strcasecmp(temp, tags[i]) == 0)
					break;
			}
			if (*tags[i] != NULLCH)
				esc = i + 1;
			else
			{
				/* Not one of these, try for a simple variable */
				if (*temp >= 'A' && *temp <= 'z')
					esc = *temp -'0';
				/* if we get here it is an unknown escape, ignore it */
				else
					continue;
			}
		}

		get_escape(cfg, temp);
		/* fprintf(stderr, "get_template type %c\n", esc + '0'); debug */
		store_escape(esc, temp);
	}

	return;
}

void get_tname(FILE * cfg, char * result)
{
	char * cp = result;
	char c;

	while ((c = tgetc(cfg)) != EOF)
	{
		if (c == '\n')
			break;
		*(cp++) = c;
	}
	*cp = NULLCH;
	/* fprintf(stderr, "get_tname returns %s %d\n", result, strlen(result)); debug */
	return;
}

void get_escape(FILE * cfg, char * result)
{
	char * cp = result;
	char c;
	int cc;

	while ((c = tgetc(cfg)) != EOF)
	{
		if (c == '\n')
			cc = -1;
		else if (c == '%' && cc == 0)
		{
			if ((c = tgetc(cfg)) == '\\')
				break;
			*(cp++) = '%';
			cc++;
		}
		*(cp++) = c;
		cc++;
	}
	*(cp - 1) = NULLCH;
	/* fprintf(stderr, "get_escape returns %s\n", result); debug */
	return;
}

void get_line(FILE * in, char * result)
{
	char * cp = result;
	int c;
	int i;

	/* skip over 5 leading spaces */
	for (i = 0; i <= 5; i++)
	{
		c = ugetc(in);
		/* Just in case some are missing */
		if (c != ' ')
			break;
	}

	/* now extract a line ending with a newline or FF */
	while (c != '\n' && c != '\f')
	{
		/* Having gone to the trouble of decoding Unicode characters we will now
			ignore them ! */
		/* if (c <= 0xFF) */
			*(cp++) = c;
		if ((c = ugetc(in)) == EOF)
		{
			--cp;
			break;
		}
	}

	*cp = NULLCH;

	/* fprintf(stderr, "get_line returns (%s) %d\n", result, strlen(result)); debug */
	return;
}

char * get_song(FILE * in)
{
	extern char * escape[];
	extern char songbuf[];
	extern char * sbp;
	char * cp;
	int c, notext;
	char temp2[SONGBUFSZ];

	/* fprintf(stderr, "get_song\n"); debug */
	*temp2 = NULLCH;

	/* Get the header */
	get_line(in, temp);

	/* if there wasn't one then we have done the last song */
	if (*temp == NULLCH)
		return temp;

	if (strcmp(temp, "Schedule Detail") != 0)
		fprintf(stderr, "Warning: this doesn't look like Schedule output.\n");

	/* The schedule name starts with "for " and ends with a file extension
		This is repeated for every song but never mind */
	get_line(in, temp);
	if (strncmp(temp, "for ", 4) != 0)
		fprintf(stderr, "Warning: this output looks damaged.\n");
	/* in practice this is never used */
	cp = temp + 4;
	while (*(cp++) != '.')
		;
	*(--cp) = NULLCH;
	store_escape(ESCAPES, temp + 4);

	/* Get the song title */
	get_line(in, temp);
	/* fprintf(stderr, "song title (%s) %d\n", temp, strlen(temp)); debug */
	store_escape(ESCAPET, temp);

	/* v1.2 Print a running tally to stderr to reassure the user */
	fprintf(stderr, "%s\n", temp);

	/* see here if this is likely to be a no text entry e.g. .ppt */
	/* Count back from the end looking for a dot */
	notext = 0;
	cp = temp + strlen(temp) - 1;
	while (*(--cp)!= '.' && cp > temp)
		;
	if ((strcmp(cp, ".ppt") == 0) || (strcmp(cp, ".pptx") == 0))
		notext = 1;

	/* Now get the song text */
	sbp = songbuf + 1;
	*sbp = NULLCH;
	get_line(in, temp);
	/* fprintf(stderr, "first tag (%s) %d\n", temp, strlen(temp)); debug */
	/* Videos are discovered because the title is the same as the first (only) line of the body */
	if (strcmp(temp, escape[ESCAPET]) == 0)
		notext = 1;

	while (!(*temp == ' ' && strlen(temp) == 1))
		/* Song ends with a line just containing a single space */
	{
		strcpy(sbp, temp);
		sbp += strlen(temp);
		*(sbp++) = '\n';
		get_line(in, temp);
	}
	*(sbp - 1) = NULLCH;

	/* Get author field */
	get_line(in, temp);
	/* Remove trailing blanks */
	cp = temp + strlen(temp) - 1;
	while (*(--cp)== ' ')
		;
	cp++;
	*cp = NULLCH;
	/* fprintf(stderr, "Author %d (%s)", strlen(temp), temp); debug */
	store_escape(ESCAPEA, temp);

	/* Get copyright field */
	get_line(in, temp);
	store_escape(ESCAPEc, temp);
	/* If we have copyright then set licence as well */
	if (*escape[ESCAPEc] == ' ')
		store_escape(ESCAPEL, "");
	else
		store_escape(ESCAPEL, escape[ESCAPEl]);

	/* Get song number field */
	get_line(in, temp);
	store_escape(ESCAPEN, temp);

	sbp = songbuf + 1;
	if (notext != 0)
		/* suppress the output */
		*sbp = NULLCH;

	/* fprintf(stderr, "get_song returns %s\n", escape[ESCAPET]); debug */
	/* fprintf(stderr, "get_song songbuf has %d \"%s\"\n", strlen(sbp), sbp); debug */

	return escape[ESCAPET];
}

char * get_stanza(void)
{
	extern char * escape[];
	extern char songbuf[];
	extern char * sbp;
	char * start;
	char * tp;
	char h;

	/* fprintf(stderr, "get_stanza sbp=%d %c\n", sbp - songbuf - 1, *sbp); debug */

	if (*sbp == NULLCH)
		return "";

	/* extract the tag at the start */
	if (is_tag(sbp) != 0)
	{
		start = sbp;
		while (*sbp != ' ' && *sbp != '\n')
			sbp++;
		h = *sbp;
		*(sbp++) = NULLCH;
		/* fprintf(stderr, "tag %s, term = %d\n", start, h); debug */
		store_escape(ESCAPEt, start);

		/* Now the tag number (or other text) if any */
		if (h == ' ')
		{
			start = sbp;
			while (*sbp != '\n')
				sbp++;
			*(sbp++) = NULLCH;
			store_escape(ESCAPEn, start);
		}
		else
			store_escape(ESCAPEn, "");
	}
	else
	{
		/* skip over blank tag */
		while (*sbp != '\n')
			sbp++;
		sbp++;
		/* fprintf(stderr, "Blank tag (%s)", sbp); debug */
		store_escape(ESCAPEt, "");
		store_escape(ESCAPEn, "");
	}

	tp = temp;

	while (is_tag(sbp) == 0)
	{
		start = sbp;
		while (*sbp != '\n' && *sbp != NULLCH)
			sbp++;
		strncpy(tp, start, sbp - start);
		tp += sbp - start;
		strcpy(tp, escape[ESCAPE_line]);
		tp += 6;
		if (*sbp == NULLCH)
			break;
		sbp++;
	}

	/* we don't want a line break on the last one */
	tp -= 6;
	*tp = NULLCH;

	store_escape(ESCAPEz, temp);

	/* fprintf(stderr, "get_stanza returns %s\n", escape[ESCAPEz]); debug */

	return escape[ESCAPEz];
}

int is_tag(char * cp)
{
	int i, tl;

	/* fprintf(stderr, "is_tag %10.10s\n", cp); debug */
	for (i = 0; *tags[i] != NULLCH; i++)
	{
		tl = strlen(tags[i]);
		if (strncasecmp(cp, tags[i], tl) != 0)
			continue;
		if (*(cp + tl) == '\n')
		{
			/* fprintf(stderr, "*is_tag returns %d\n", i + 1); debug */
			return i + 1;
		}
		/* the next character must be a space */
		if (*(cp + tl) != ' ')
			continue;
		/* and the one after numeric or ( */
		if (!isdigit(*(cp + tl + 1)) && *(cp + tl + 1) != '(')
			continue;
		/* fprintf(stderr, "is_tag returns %d\n", i + 1); debug */
		return i + 1;
	}
	return 0;
}

void put_schedule(FILE * out, FILE * in)
{
	extern char * escape[];
	extern int inschd, insong;

	inschd = FALSE;
	insong = FALSE;

	/* fprintf(stderr, "put_schedule\n"); debug */

	put_text(out, escape[ESCAPE_schedule], in);
	return;
}

/* note - this subroutine is recursive. Care must be taken with any
	static variables that it uses directly or indirectly */
void put_text(FILE * out, char * string, FILE * in)
{
	extern char * escape[];
	extern int inschd, insong;

	int i;
	int c;
	char * cp;

	/* fprintf(stderr, "put_text(%s) %d %d\n", string, inschd, insong); debug */

	cp = string;
	for (; (c = *cp) != NULLCH; cp++)
	{
		if (c != '%')
			fputc(c, out);
		else
		{
			c = *(++cp);
			/* fprintf(stderr, "escape(%c)\n", c); debug */
			if (c == '%')
				fputc('%', out);
			else if (c == 'C' && !inschd)
			{ /* Schedule content escape */
				inschd=TRUE;
				while (strcmp(get_song(in), "") != 0)
				{
					/* PowerPoint, Images, Videos etc. have null content */
					if (*sbp != NULLCH)
						put_text(out, escape[ESCAPE_song], in);
				}
				inschd=FALSE;
			}
			else if (c == 's' && !insong)
			{ /* Song content escape */
				insong=TRUE;
				while (strcmp(get_stanza(), "") != 0)
				{
					/* see if we have a special template for this tag */
					for (i = 0; *tags[i] != NULLCH; i++)
					{
						if (strcasecmp(escape[ESCAPEt], tags[i]) == 0)
							break;
					}
					if (*tags[i] == NULLCH)
						i = 0;
					else
						i++;
					if (*(escape[ESCAPE_stanza + i]) == NULLCH)
						i = 0;
					put_text(out, escape[ESCAPE_stanza + i], in);
				}
				insong=FALSE;
			}
			else if (c > '0' && c <= 'z' && escape[c - '0'] != (char *)NULL)
				put_text(out, escape[c - '0'], in);
/* this has the effect of outputting null for any unknown or undefined
	escape characters. Beware of lone % characters in templates.
*/
		}
	}
	return;
}

void store_escape(int n, char * string)
{
	int l;

	/* fprintf(stderr, "ESCAPE %c = \"%s\"\n", n + '0', string); debug */
	if (escape[n] != (char *)NULL)
		free(escape[n]);

	l = strlen(string);
	escape[n] = (char *)malloc(sizeof(char) * l + 1);
	strcpy(escape[n], string);

	return;
}
